
/*
 * auto-generated slotsig_4.h header file
 *
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 * This file provides support for signals and slots using 4 parameter(s)
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#ifndef __SLOTSIG_4_H__
#define __SLOTSIG_4_H__ 1

#include <stdexcept>

#include "slotsig_tools.h"
#include "slotsig_bases.h"
#include "slotsig_policies.h"

namespace SlotSig {

/*!
  \brief Base virtual class for slots receiving 4 parameters.

  This class if a functor : its operator() runs the stored slot.
*/

template <typename _RETURN_TYPE, typename _ARG0_TYPE, typename _ARG1_TYPE, typename _ARG2_TYPE, typename _ARG3_TYPE>
class Slot4 : public SlotBase
{ public:
    virtual _RETURN_TYPE operator() (typename Type2Param<_ARG0_TYPE>::ParamType arg0, typename Type2Param<_ARG1_TYPE>::ParamType arg1, typename Type2Param<_ARG2_TYPE>::ParamType arg2, typename Type2Param<_ARG3_TYPE>::ParamType arg3) const = 0 ;
    virtual bool equals(Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* sl) const = 0 ;
    virtual SlotsSetBase* in_class() const
    {
      return 0 ;
    }
} ; // class Slot4<>

/*!
  \brief Class to hold slots which are global functions receiving 4 parameters.
*/
template <typename _RETURN_TYPE, typename _ARG0_TYPE, typename _ARG1_TYPE, typename _ARG2_TYPE, typename _ARG3_TYPE>
class SlotGlobal4 : public Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>
{
  public:
    typedef _RETURN_TYPE (*global_slot_t) (_ARG0_TYPE arg0, _ARG1_TYPE arg1, _ARG2_TYPE arg2, _ARG3_TYPE arg3) ;

    SlotGlobal4 (global_slot_t gs) :
      global_slot(gs)
    { }

    virtual bool equals(Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* sl) const
    {
      SlotGlobal4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* sg =
        static_cast<SlotGlobal4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*>(sl) ;
      if ( sg != 0 )
        return sg->matches((void*)0, global_slot) ;
      else
        return false ;
    }

    virtual _RETURN_TYPE operator()(typename Type2Param<_ARG0_TYPE>::ParamType arg0, typename Type2Param<_ARG1_TYPE>::ParamType arg1, typename Type2Param<_ARG2_TYPE>::ParamType arg2, typename Type2Param<_ARG3_TYPE>::ParamType arg3) const
    {
      return global_slot(arg0, arg1, arg2, arg3) ;
    }

    bool matches(void* v, global_slot_t gs) const
    {
      return ((v == 0) && (gs == global_slot)) ;
    }

    const global_slot_t& the_slot() const
    {
      return global_slot ;
    }

  private:
    global_slot_t global_slot ;
} ; // class SlotGlobal4<>

/*!
  \brief Class to hold slots which are class's methods receiving 4 parameters.
*/
template <class _C, typename _RETURN_TYPE, typename _METHOD_TYPE, typename _ARG0_TYPE, typename _ARG1_TYPE, typename _ARG2_TYPE, typename _ARG3_TYPE>
class SlotInClass4 : public Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>
{
  public:
    SlotInClass4 (_C* c, _METHOD_TYPE m) :
      Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>(),
      _in_class(c),
      _method(m)
    { }

    virtual _RETURN_TYPE operator()(typename Type2Param<_ARG0_TYPE>::ParamType arg0, typename Type2Param<_ARG1_TYPE>::ParamType arg1, typename Type2Param<_ARG2_TYPE>::ParamType arg2, typename Type2Param<_ARG3_TYPE>::ParamType arg3) const
    {
      return (_in_class->*_method)(arg0, arg1, arg2, arg3) ;
    }

    virtual bool equals(Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* sl) const
    {
      SlotInClass4<_C, _RETURN_TYPE, _METHOD_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* sic =
        static_cast<SlotInClass4<_C, _RETURN_TYPE, _METHOD_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*>(sl) ;
      if ( sic != 0 )
        return sic->matches(_in_class, _method) ;
      else
        return false ;
    }

    bool matches(_C* c, _METHOD_TYPE m) const
    {
      return ((_in_class == c) && (_method == m)) ;
    }

    virtual SlotsSetBase* in_class() const
    {
      return dynamic_cast<SlotsSetBase*>(_in_class) ;
    }

    const _METHOD_TYPE& the_slot() const
    {
      return _method ;
    }

  protected:
    _C* _in_class ;
    _METHOD_TYPE _method ;
} ; // class SlotInClass4

/*!
  \brief The class used to send a signal with 4 parameters.
*/
template <typename _RETURN_TYPE, typename _ARG0_TYPE, typename _ARG1_TYPE, typename _ARG2_TYPE, typename _ARG3_TYPE,
          template <typename _T> class _STORAGE = SimpleStorage>
class Sig4 : public SignalBase,
                protected _STORAGE<Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*>
{
  public:
    typedef _RETURN_TYPE (*global_slot_t) (_ARG0_TYPE arg0, _ARG1_TYPE arg1, _ARG2_TYPE arg2, _ARG3_TYPE arg3) ;
    typedef _STORAGE<Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*/*,
                     equal_slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*/> Storage ;

    template <class _IN_CLASS>
    struct MethodTypes
    {
      typedef _RETURN_TYPE (_IN_CLASS::*method_t) (_ARG0_TYPE arg0, _ARG1_TYPE arg1, _ARG2_TYPE arg2, _ARG3_TYPE arg3) ;
      typedef _RETURN_TYPE (_IN_CLASS::*method_const_t) (_ARG0_TYPE arg0, _ARG1_TYPE arg1, _ARG2_TYPE arg2, _ARG3_TYPE arg3) const ;
    } ;

    explicit Sig4(bool ensure_unique = true) :
      SignalBase(),
      _STORAGE<Slot4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*>(),
      _is_emitting(false),
      _ensure_unique(ensure_unique)
    { }

    ~Sig4()
    {
      if ( _is_emitting )
        throw std::logic_error("SlotSig4 : should not delete a signal while emitting !") ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
        {
          SlotsSetBase* ssb = (*i)->in_class() ;
          if ( ssb != 0 )
            ssb->signalDeleted(this) ;
          delete (*i) ;
        }
        ++i ;
      }
    }

    SlotSigErrors disconnect_all()
    {
      if ( _is_emitting )
        return IsEmitting ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
        {
          SlotsSetBase* ssb = (*i)->in_class() ;
          if ( ssb != 0 )
            ssb->disconnectFrom(this) ;
          delete (*i) ;
          i = erase(i) ;
        }
        else
          ++i ;
      }
    }

    bool ensure_unique() const
    {
      return ensure_unique ;
    }

    SlotSigErrors set_ensure_unique(bool ensure)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( ensure == _ensure_unique )
        return NoError ;
      _ensure_unique = ensure ;
      return NoError ;
    }

    SlotSigErrors enforce_unique()
    {
      if ( _is_emitting )
        return IsEmitting ;
      typename Storage::iterator e = end() ;
      typename Storage::iterator i = begin() ;
      Storage tmp_store ;
      const unsigned nb_conn = nb_connected() ;
      unsigned nb_done = 0 ;
      while ( (i != e) && (nb_done < nb_conn) )
      {
        if ( (*i) != 0 )
        {
          typename Storage::iterator i2 =
            tmp_store.find_equal(*i) ;
          if ( i2 == tmp_store.end() )
            tmp_store.insert(*i) ;
          ++nb_done ;
        }
        ++i ;
      }
      clear() ;
      e = tmp_store.end() ;
      i = tmp_store.begin() ;
      while ( i != e )
        insert(*(i++)) ;
      return NoError ;
    }

    unsigned nb_connected() const
    {
      return Storage::stored() ;
    }

    size_t capacity() const
    {
      return Storage::capacity() ;
    }

    SlotSigErrors connect (global_slot_t f)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( f == 0 )
        return NullGlobalFct ;
      if ( _ensure_unique )
      {
        typename Storage::iterator i =
          find((void*)0, f, (SlotGlobal4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*)0) ;
        if ( i != end() )
          return KnownGlobalFct ;
      }
      SlotGlobal4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* slot =
        new SlotGlobal4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>(f) ;
      if ( insert(slot) )
        return NoError ;
      else
      {
        delete slot ;
        return KnownGlobalFct ;
      }
    }

    template <class _C>
    SlotSigErrors connect (_C* in_class)
    {
      return connect (in_class, &_C::operator()) ;
    }

    template <class _C>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C::*method)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE))
    {
      return connectGen(in_class, method) ;
    }

    template <class _C>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C::*method)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE) const)
    {
      return connectGen(in_class, method) ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C_PARENT::*method)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE))
    {
      _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return connectGen (parent_class, method) ;
      else
        return DynCastFailed ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C_PARENT::*method)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE) const)
    {
      _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return connectGen (parent_class, method) ;
      else
        return DynCastFailed ;
    }

    SlotSigErrors disconnect (global_slot_t f)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( f == 0 )
        return NullGlobalFct ;
      typename Storage::iterator i =
        find((void*)0, f, (SlotGlobal4<_RETURN_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*)0) ;
      if ( i == end() )
        return UnknownGlobalFct ;
      delete *i ;
      erase(i) ;
      return NoError ;
    }

    template <class _C>
    SlotSigErrors disconnect (_C* in_class)
    {
      return disconnect(in_class, &_C::operator()) ;
    }

    template <class _C>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C::*methode)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE))
    {
      return disconnectGen(in_class, methode) ;
    }

    template <class _C>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C::*methode)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE) const)
    {
      return disconnectGen(in_class, methode) ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C_PARENT::*methode)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE))
    { _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return disconnectGen (parent_class, methode) ;
      else
        return DynCastFailed ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C_PARENT::*methode)(_ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE) const)
    { _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return disconnectGen (parent_class, methode) ;
      else
        return DynCastFailed ;
    }

    bool is_emitting() const
    {
      return _is_emitting ;
    }

    SlotSigErrors run(typename Type2Param<_ARG0_TYPE>::ParamType arg0, typename Type2Param<_ARG1_TYPE>::ParamType arg1, typename Type2Param<_ARG2_TYPE>::ParamType arg2, typename Type2Param<_ARG3_TYPE>::ParamType arg3) const
    {
      if ( _is_emitting )
        return IsEmitting ;
      _is_emitting = true ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          (*(*i))(arg0, arg1, arg2, arg3) ;
        ++i ;
      }
      _is_emitting = false ;
      return NoError ;
    }

    template <class _ACCUMULATOR>
    SlotSigErrors run(_ACCUMULATOR& accum, typename Type2Param<_ARG0_TYPE>::ParamType arg0, typename Type2Param<_ARG1_TYPE>::ParamType arg1, typename Type2Param<_ARG2_TYPE>::ParamType arg2, typename Type2Param<_ARG3_TYPE>::ParamType arg3) const
    {
      if ( _is_emitting )
        return IsEmitting ;
      _is_emitting = true ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          accum((*(*i))(arg0, arg1, arg2, arg3)) ;
        ++i ;
      }
      _is_emitting = false ;
      return NoError ;
    }

    template <class _PREDICATE>
    SlotSigErrors run_if(_PREDICATE& pred, typename Type2Param<_ARG0_TYPE>::ParamType arg0, typename Type2Param<_ARG1_TYPE>::ParamType arg1, typename Type2Param<_ARG2_TYPE>::ParamType arg2, typename Type2Param<_ARG3_TYPE>::ParamType arg3) const
    {
      if ( _is_emitting )
        return IsEmitting ;
      _is_emitting = true ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          if ( ! pred((*(*i))(arg0, arg1, arg2, arg3)) )
          {
            _is_emitting = false ;
            return EmitInterrupted ;
          }
        ++i ;
      }
      _is_emitting = false ;
      return NoError ;
    }

  protected:
    virtual void slotsSetDeleted (SlotsSetBase* ssb)
    {
      if ( ssb == 0 )
        return ;
      typename Storage::iterator i = begin() ;
      while ( i != end() )
      {
        if ( (*i) != 0 )
        {
          if ( (*i)->in_class() == ssb )
          {
            delete *i ;
            i = erase(i) ;
          }
          else
            ++i ;
        }
        else
          ++i ;
      }
    }

  private:
    template <class _C, typename _METHOD_TYPE>
    SlotSigErrors connectGen (_C* in_class,
                     _METHOD_TYPE method)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( in_class == 0 )
        return NullInstance ;
      if ( _ensure_unique )
      {
        typename Storage::iterator i =
          find(in_class, method,
          (SlotInClass4<_C, _RETURN_TYPE, _METHOD_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*)0) ;
        if ( i != end() )
          return KnownSlot ;
      }
      SlotInClass4<_C, _RETURN_TYPE, _METHOD_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>* slot =
        new SlotInClass4<_C, _RETURN_TYPE, _METHOD_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>(in_class, method) ;
      if ( insert(slot) )
      {
        SlotsSetBase* ssb = dynamic_cast<SlotsSetBase*>(in_class) ;
        if ( ssb != 0 )
          ssb->connectFrom(this) ;
        return NoError ;
      }
      else
      {
        delete slot ;
        return KnownSlot ;
      }
    }

    template <class _C, typename _METHOD_TYPE>
    SlotSigErrors disconnectGen (_C* in_class,
                        _METHOD_TYPE method)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( in_class == 0 )
        return NullInstance ;
      typename Storage::iterator i =
        find(in_class, method,
        (SlotInClass4<_C, _RETURN_TYPE, _METHOD_TYPE, _ARG0_TYPE, _ARG1_TYPE, _ARG2_TYPE, _ARG3_TYPE>*)0) ;
      if ( i == end() )
        return UnknownSlot ;
      SlotsSetBase* ssb = dynamic_cast<SlotsSetBase*>(in_class) ;
      if ( ssb != 0 )
        ssb->disconnectFrom(this) ;
      delete *i ;
      erase(i) ;
      return NoError ;
    }

    mutable bool _is_emitting ;
    bool _ensure_unique ;
} ; // class Sig4<>

} // namespace SlotSig

#endif // __SLOTSIG_4_H__
