
/*
 * slotsig_base.h header file
 *
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 * This file provides base support for signals and slots, and should be
 * compiled with your own program.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <iostream>

#include "slotsig_bases.h"

namespace SlotSig
{

const char* error2str(SlotSigErrors error)
{
  switch ( error )
  {
    case NoError          : return "NoError" ;
    case NullGlobalFct    : return "NullGlobalFct" ;
    case KnownGlobalFct   : return "KnownGlobalFct" ;
    case UnknownGlobalFct : return "UnknownGlobalFct" ;
    case DynCastFailed    : return "DynCastFailed" ;
    case NullInstance     : return "NullInstance" ;
    case KnownSlot        : return "KnownSlot" ;
    case UnknownSlot      : return "UnknownSlot" ;
    case IsEmitting       : return "IsEmitting" ;
    case EmitInterrupted  : return "EmitInterrupted" ;
    case NullSignal       : return "NullSignal" ;
  }
  return "UnknownError" ;
}

SlotBase::~SlotBase()
{ }

SignalBase::SignalBase()
{ }

SignalBase::~SignalBase()
{ }

SlotsSetBase::SlotsSetBase() :
  signals_map()
{ }

SlotsSetBase::~SlotsSetBase()
{
  const signals_map_t::const_iterator end = signals_map.end() ;
  signals_map_t::const_iterator iter = signals_map.begin() ;
  while ( iter != end )
    (iter++)->first->slotsSetDeleted(this) ;
}

void SlotsSetBase::disconnectAll()
{
  const signals_map_t::const_iterator end = signals_map.end() ;
  signals_map_t::const_iterator iter = signals_map.begin() ;
  while ( iter != end )
    (iter++)->first->slotsSetDeleted(this) ;
  signals_map.clear() ;
}

void SlotsSetBase::connectFrom(SignalBase* sig)
{
  if ( sig == 0 )
    return ;
  ++(signals_map[sig]) ;
}

void SlotsSetBase::disconnectFrom(SignalBase* sig)
{
  if ( sig == 0 )
    return ;
  signals_map_t::iterator iter = signals_map.find(sig) ;
  if ( iter != signals_map.end() )
  {
    if ( iter->second == 1 )
      signals_map.erase(iter) ;
    else
      --(iter->second) ;
  }
}

void SlotsSetBase::signalDeleted(SignalBase* sig)
{
  if ( sig == 0 )
    return ;
  signals_map.erase(sig) ;
}

} // namespace SlotSig
