
/*
 * auto-generated slotsig_@NB@.h header file
 *
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 * This file provides support for signals and slots using @NB@ parameter(s)
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#ifndef __SLOTSIG_@NB@_H__
#define __SLOTSIG_@NB@_H__ 1

#include <stdexcept>

#include "slotsig_tools.h"
#include "slotsig_bases.h"
#include "slotsig_policies.h"

namespace SlotSig {

/*!
  \brief Base virtual class for slots receiving @NB@ parameters.

  This class if a functor : its operator() runs the stored slot.
*/

template <typename _RETURN_TYPE@@ @TPL_PARAMS@>
class Slot@NB@ : public SlotBase
{ public:
    virtual _RETURN_TYPE operator() (@SLT_PARAMS@) const = 0 ;
    virtual bool equals(Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>* sl) const = 0 ;
    virtual SlotsSetBase* in_class() const
    {
      return 0 ;
    }
} ; // class Slot@NB@<>

/*!
  \brief Class to hold slots which are global functions receiving @NB@ parameters.
*/
template <typename _RETURN_TYPE@@ @TPL_PARAMS@>
class SlotGlobal@NB@ : public Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>
{
  public:
    typedef _RETURN_TYPE (*global_slot_t) (@FCT_PARAMS@) ;

    SlotGlobal@NB@ (global_slot_t gs) :
      global_slot(gs)
    { }

    virtual bool equals(Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>* sl) const
    {
      SlotGlobal@NB@<_RETURN_TYPE@@ @TPL_ARGS@>* sg =
        static_cast<SlotGlobal@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*>(sl) ;
      if ( sg != 0 )
        return sg->matches((void*)0, global_slot) ;
      else
        return false ;
    }

    virtual _RETURN_TYPE operator()(@SLT_PARAMS@) const
    {
      return global_slot(@FCT_ARGS@) ;
    }

    bool matches(void* v, global_slot_t gs) const
    {
      return ((v == 0) && (gs == global_slot)) ;
    }

    const global_slot_t& the_slot() const
    {
      return global_slot ;
    }

  private:
    global_slot_t global_slot ;
} ; // class SlotGlobal@NB@<>

/*!
  \brief Class to hold slots which are class's methods receiving @NB@ parameters.
*/
template <class _C, typename _RETURN_TYPE, typename _METHOD_TYPE@@ @TPL_PARAMS@>
class SlotInClass@NB@ : public Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>
{
  public:
    SlotInClass@NB@ (_C* c, _METHOD_TYPE m) :
      Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>(),
      _in_class(c),
      _method(m)
    { }

    virtual _RETURN_TYPE operator()(@SLT_PARAMS@) const
    {
      return (_in_class->*_method)(@FCT_ARGS@) ;
    }

    virtual bool equals(Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>* sl) const
    {
      SlotInClass@NB@<_C, _RETURN_TYPE, _METHOD_TYPE@@ @TPL_ARGS@>* sic =
        static_cast<SlotInClass@NB@<_C, _RETURN_TYPE, _METHOD_TYPE@@ @TPL_ARGS@>*>(sl) ;
      if ( sic != 0 )
        return sic->matches(_in_class, _method) ;
      else
        return false ;
    }

    bool matches(_C* c, _METHOD_TYPE m) const
    {
      return ((_in_class == c) && (_method == m)) ;
    }

    virtual SlotsSetBase* in_class() const
    {
      return dynamic_cast<SlotsSetBase*>(_in_class) ;
    }

    const _METHOD_TYPE& the_slot() const
    {
      return _method ;
    }

  protected:
    _C* _in_class ;
    _METHOD_TYPE _method ;
} ; // class SlotInClass@NB@

/*!
  \brief The class used to send a signal with @NB@ parameters.
*/
template <typename _RETURN_TYPE@@ @TPL_PARAMS@,
          template <typename _T> class _STORAGE = SimpleStorage>
class Sig@NB@ : public SignalBase,
                protected _STORAGE<Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*>
{
  public:
    typedef _RETURN_TYPE (*global_slot_t) (@FCT_PARAMS@) ;
    typedef _STORAGE<Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*/*,
                     equal_slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*/> Storage ;

    template <class _IN_CLASS>
    struct MethodTypes
    {
      typedef _RETURN_TYPE (_IN_CLASS::*method_t) (@FCT_PARAMS@) ;
      typedef _RETURN_TYPE (_IN_CLASS::*method_const_t) (@FCT_PARAMS@) const ;
    } ;

    explicit Sig@NB@(bool ensure_unique = true) :
      SignalBase(),
      _STORAGE<Slot@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*>(),
      _is_emitting(false),
      _ensure_unique(ensure_unique)
    { }

    ~Sig@NB@()
    {
      if ( _is_emitting )
        throw std::logic_error("SlotSig@NB@ : should not delete a signal while emitting !") ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
        {
          SlotsSetBase* ssb = (*i)->in_class() ;
          if ( ssb != 0 )
            ssb->signalDeleted(this) ;
          delete (*i) ;
        }
        ++i ;
      }
    }

    SlotSigErrors disconnect_all()
    {
      if ( _is_emitting )
        return IsEmitting ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
        {
          SlotsSetBase* ssb = (*i)->in_class() ;
          if ( ssb != 0 )
            ssb->disconnectFrom(this) ;
          delete (*i) ;
          i = erase(i) ;
        }
        else
          ++i ;
      }
    }

    bool ensure_unique() const
    {
      return ensure_unique ;
    }

    SlotSigErrors set_ensure_unique(bool ensure)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( ensure == _ensure_unique )
        return NoError ;
      _ensure_unique = ensure ;
      return NoError ;
    }

    SlotSigErrors enforce_unique()
    {
      if ( _is_emitting )
        return IsEmitting ;
      typename Storage::iterator e = end() ;
      typename Storage::iterator i = begin() ;
      Storage tmp_store ;
      const unsigned nb_conn = nb_connected() ;
      unsigned nb_done = 0 ;
      while ( (i != e) && (nb_done < nb_conn) )
      {
        if ( (*i) != 0 )
        {
          typename Storage::iterator i2 =
            tmp_store.find_equal(*i) ;
          if ( i2 == tmp_store.end() )
            tmp_store.insert(*i) ;
          ++nb_done ;
        }
        ++i ;
      }
      clear() ;
      e = tmp_store.end() ;
      i = tmp_store.begin() ;
      while ( i != e )
        insert(*(i++)) ;
      return NoError ;
    }

    unsigned nb_connected() const
    {
      return Storage::stored() ;
    }

    size_t capacity() const
    {
      return Storage::capacity() ;
    }

    SlotSigErrors connect (global_slot_t f)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( f == 0 )
        return NullGlobalFct ;
      if ( _ensure_unique )
      {
        typename Storage::iterator i =
          find((void*)0, f, (SlotGlobal@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*)0) ;
        if ( i != end() )
          return KnownGlobalFct ;
      }
      SlotGlobal@NB@<_RETURN_TYPE@@ @TPL_ARGS@>* slot =
        new SlotGlobal@NB@<_RETURN_TYPE@@ @TPL_ARGS@>(f) ;
      if ( insert(slot) )
        return NoError ;
      else
      {
        delete slot ;
        return KnownGlobalFct ;
      }
    }

    template <class _C>
    SlotSigErrors connect (_C* in_class)
    {
      return connect (in_class, &_C::operator()) ;
    }

    template <class _C>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C::*method)(@TPL_ARGS@))
    {
      return connectGen(in_class, method) ;
    }

    template <class _C>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C::*method)(@TPL_ARGS@) const)
    {
      return connectGen(in_class, method) ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C_PARENT::*method)(@TPL_ARGS@))
    {
      _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return connectGen (parent_class, method) ;
      else
        return DynCastFailed ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors connect (_C* in_class,
                  _RETURN_TYPE (_C_PARENT::*method)(@TPL_ARGS@) const)
    {
      _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return connectGen (parent_class, method) ;
      else
        return DynCastFailed ;
    }

    SlotSigErrors disconnect (global_slot_t f)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( f == 0 )
        return NullGlobalFct ;
      typename Storage::iterator i =
        find((void*)0, f, (SlotGlobal@NB@<_RETURN_TYPE@@ @TPL_ARGS@>*)0) ;
      if ( i == end() )
        return UnknownGlobalFct ;
      delete *i ;
      erase(i) ;
      return NoError ;
    }

    template <class _C>
    SlotSigErrors disconnect (_C* in_class)
    {
      return disconnect(in_class, &_C::operator()) ;
    }

    template <class _C>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C::*methode)(@TPL_ARGS@))
    {
      return disconnectGen(in_class, methode) ;
    }

    template <class _C>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C::*methode)(@TPL_ARGS@) const)
    {
      return disconnectGen(in_class, methode) ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C_PARENT::*methode)(@TPL_ARGS@))
    { _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return disconnectGen (parent_class, methode) ;
      else
        return DynCastFailed ;
    }

    template <class _C, class _C_PARENT>
    SlotSigErrors disconnect (_C* in_class,
                     _RETURN_TYPE (_C_PARENT::*methode)(@TPL_ARGS@) const)
    { _C_PARENT* parent_class = dynamic_cast<_C_PARENT*>(in_class) ;
      if ( parent_class != 0 )
        return disconnectGen (parent_class, methode) ;
      else
        return DynCastFailed ;
    }

    bool is_emitting() const
    {
      return _is_emitting ;
    }

    SlotSigErrors run(@SLT_PARAMS@) const
    {
      if ( _is_emitting )
        return IsEmitting ;
      _is_emitting = true ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          (*(*i))(@FCT_ARGS@) ;
        ++i ;
      }
      _is_emitting = false ;
      return NoError ;
    }

    template <class _ACCUMULATOR>
    SlotSigErrors run(_ACCUMULATOR& accum@@ @SLT_PARAMS@) const
    {
      if ( _is_emitting )
        return IsEmitting ;
      _is_emitting = true ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          accum((*(*i))(@FCT_ARGS@)) ;
        ++i ;
      }
      _is_emitting = false ;
      return NoError ;
    }

    template <class _PREDICATE>
    SlotSigErrors run_if(_PREDICATE& pred@@ @SLT_PARAMS@) const
    {
      if ( _is_emitting )
        return IsEmitting ;
      _is_emitting = true ;
      const typename Storage::const_iterator e = end() ;
      typename Storage::const_iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          if ( ! pred((*(*i))(@FCT_ARGS@)) )
          {
            _is_emitting = false ;
            return EmitInterrupted ;
          }
        ++i ;
      }
      _is_emitting = false ;
      return NoError ;
    }

  protected:
    virtual void slotsSetDeleted (SlotsSetBase* ssb)
    {
      if ( ssb == 0 )
        return ;
      typename Storage::iterator i = begin() ;
      while ( i != end() )
      {
        if ( (*i) != 0 )
        {
          if ( (*i)->in_class() == ssb )
          {
            delete *i ;
            i = erase(i) ;
          }
          else
            ++i ;
        }
        else
          ++i ;
      }
    }

  private:
    template <class _C, typename _METHOD_TYPE>
    SlotSigErrors connectGen (_C* in_class,
                     _METHOD_TYPE method)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( in_class == 0 )
        return NullInstance ;
      if ( _ensure_unique )
      {
        typename Storage::iterator i =
          find(in_class, method,
          (SlotInClass@NB@<_C, _RETURN_TYPE, _METHOD_TYPE@@ @TPL_ARGS@>*)0) ;
        if ( i != end() )
          return KnownSlot ;
      }
      SlotInClass@NB@<_C, _RETURN_TYPE, _METHOD_TYPE@@ @TPL_ARGS@>* slot =
        new SlotInClass@NB@<_C, _RETURN_TYPE, _METHOD_TYPE@@ @TPL_ARGS@>(in_class, method) ;
      if ( insert(slot) )
      {
        SlotsSetBase* ssb = dynamic_cast<SlotsSetBase*>(in_class) ;
        if ( ssb != 0 )
          ssb->connectFrom(this) ;
        return NoError ;
      }
      else
      {
        delete slot ;
        return KnownSlot ;
      }
    }

    template <class _C, typename _METHOD_TYPE>
    SlotSigErrors disconnectGen (_C* in_class,
                        _METHOD_TYPE method)
    {
      if ( _is_emitting )
        return IsEmitting ;
      if ( in_class == 0 )
        return NullInstance ;
      typename Storage::iterator i =
        find(in_class, method,
        (SlotInClass@NB@<_C, _RETURN_TYPE, _METHOD_TYPE@@ @TPL_ARGS@>*)0) ;
      if ( i == end() )
        return UnknownSlot ;
      SlotsSetBase* ssb = dynamic_cast<SlotsSetBase*>(in_class) ;
      if ( ssb != 0 )
        ssb->disconnectFrom(this) ;
      delete *i ;
      erase(i) ;
      return NoError ;
    }

    mutable bool _is_emitting ;
    bool _ensure_unique ;
} ; // class Sig@NB@<>

} // namespace SlotSig

#endif // __SLOTSIG_@NB@_H__
