
// Compile command :
// g++ -Wall -o test_@NB@.x -I.. -lstdc++ ../slotsig/slotsig_bases.cpp test_@NB@.cpp
#include <iostream>
#include <string>
#include <slotsig/slotsig_@NB@.h>

class C1 : public SlotSig::SlotsSetBase
{
  public:
    C1() : SlotSig::SlotsSetBase() { }

    std::string m1(@FCT_PARAMS@)
    {
      std::string s = @FCT_CONCAT@ ;
      std::cout << "C1::m1 (" << s << ")\n" ;
      return "C1::m1" ;
    }

    std::string m2(@FCT_PARAMS@) const
    {
      std::string s = @FCT_CONCAT@ ;
      std::cout << "C1::m2 (" << s << ")\n" ;
      return "C1::m2" ;
    }

    virtual std::string vm1(@FCT_PARAMS@)
    {
      std::string s = @FCT_CONCAT@ ;
      std::cout << "C1::vm1 (" << s << ")\n" ;
      return "C1::vm1" ;
    }

    virtual std::string vm2(@FCT_PARAMS@) const
    {
      std::string s = @FCT_CONCAT@ ;
      std::cout << "C1::vm2 (" << s << ")\n" ;
      return "C1::vm2" ;
    }
} ; // C1

class C2 : public C1
{
  public:
    C2() : C1() { }

    virtual std::string vm1(@FCT_PARAMS@)
    {
      std::string s = @FCT_CONCAT@ ;
      std::cout << "C2::vm1 (" << s << ")\n" ;
      return "C2::vm1" ;
    }

    virtual std::string vm2(@FCT_PARAMS@) const
    {
      std::string s = @FCT_CONCAT@ ;
      std::cout << "C2::vm2 (" << s << ")\n" ;
      return "C2::vm2" ;
    }
} ; // C2

std::string global_slot(@FCT_PARAMS@)
{
  std::string s = @FCT_CONCAT@ ;
  std::cout << "::global_slot (" << s << ")\n" ;
  return "::global_slot" ;
}

struct Accum
{
  Accum() : str("") { }

  void operator() (const std::string& s)
  { str += (s + " ") ; }

  std::string str ;
} ; // Accum

struct Pred
{
  Pred() : cpt(0) { }

  bool operator() (const std::string&)
  { return ( ((++cpt) % 3) != 0 ) ; }

  unsigned cpt ;
} ; // Pred

int main()
{
  std::cout << "--- Test program START ---\n" ;

  C1* c1 = new C1 ;
  C2* c2 = new C2 ;
  typedef SlotSig::Sig@NB@<std::string@@ @TPL_ARGS@> SigType ;
  SigType* sig1 = new SigType(false) ;
  SigType* sig2 = new SigType ;
  Accum accum ;
  Pred pred ;

  std::cout << "--- Connecting all methods from c1 to sig1...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (global_slot)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::vm1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::vm2)) << "\n" ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Re-connecting all methods from c1 to sig1...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (global_slot)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::vm1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c1, &C1::vm2)) << "\n" ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Ensuring uniqueness of slots in sig1...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->set_ensure_unique(true)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->enforce_unique()) << "\n" ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Connecting all methods from c2 to sig1...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::vm1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::vm2)) << "\n" ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Re-connecting all methods from c2 to sig1...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::vm1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->connect (c2, &C2::vm2)) << "\n" ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Connecting all methods from c1 and c2 to sig2...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (global_slot)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c1, &C1::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c1, &C1::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c1, &C1::vm1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c1, &C1::vm2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c2, &C2::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c2, &C2::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c2, &C2::vm1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->connect (c2, &C2::vm2)) << "\n" ;
  std::cout << "--- Emitting sig2 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig2 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Disconnecting global_slot, c1->m1, c1->vm2, c2->m2 and c2->vm1 from sig2...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->disconnect(global_slot)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->disconnect(c1, &C1::m1)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->disconnect(c1, &C1::vm2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->disconnect(c2, &C2::m2)) << "\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->disconnect(c2, &C2::vm1)) << "\n" ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig2 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Deleting c1...\n" ;
  delete c1 ;
  c1 = 0 ;
  std::cout << "--- Emitting sig1 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig1 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig1->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig1 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig1->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig2 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Deleting sig1...\n" ;
  delete sig1 ;
  sig1 = 0 ;
  std::cout << "--- Emitting sig2 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig2 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Deleting c2...\n" ;
  delete c2 ;
  std::cout << "--- Emitting sig2 ##...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(@FCT_ARGS_##@)) << "\n" ;
  std::cout << "--- Emitting sig2 ## with accumulator...\n" ;
  accum.str = "" ;
  std::cout << "  " << SlotSig::error2str(sig2->run(accum@@ @FCT_ARGS_##@)) << "\n" ;
  std::cout << "accum.str = " << accum.str << "\n" ;
  std::cout << "--- Emitting sig2 ## with predicate (run_if)...\n" ;
  std::cout << "  " << SlotSig::error2str(sig2->run_if(pred@@ @FCT_ARGS_##@)) << "\n" ;

  std::cout << "--- Deleting sig2...\n" ;
  delete sig2 ;

  std::cout << "--- Test program END ----\n" ;
  return 0 ;
}
